// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
#include "sudoku.h"
#include <stdlib.h>
#include <math.h>

#ifdef _DEBUG
#include <windows.h>
#include <stdio.h>
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
static void GenGame();
static int  NewCellNumber(int   row, int   col);
static bool FndGameFixd  (int& frow, int& fcol);
static void RstGame();
static void RstGameCell(int row1, int col1, bool fix);

extern "C" int SolvePuzzle(char* game, int maxSolutions, int* score);

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
GameCell gameCell[SDSIZESQ][SDSIZESQ]; // row, col
bool     gameErrs[SDSIZESQ][SDSIZESQ]; // row, col
GameType gameType   =   gameEasy;
int      gameSeed   =   0;
bool     gameWon    =   false;
int      gameRow    =   0;
int      gameCol    =   0;
int      gameDiff[] = { 2, 9, 18, 27, 45, 100 };
int      gameLevel  =   9;
int      gameSolves =   0;
bool     gameGen    =   false;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
void NewGame()
{
	if( !gameGen )
	{
		gameSeed++;
		BegGame();
	}
}

// ----------------------------------------------------------------------------
void ClrGame()
{
	for( int row = 0; row < SDSIZESQ; row++ )
	{
		for( int col = 0; col < SDSIZESQ; col++ )
		{
			gameCell[row][col].nbr = 0;
			gameCell[row][col].usr = 0;
			gameCell[row][col].fix = false;
			gameCell[row][col].sho = false;
			gameErrs[row][col]     = false;
		}
	}
}

// ----------------------------------------------------------------------------
void SetGame(int seed, char* user, char* grid, char* give)
{
	gameSeed = seed;
	ClrGame();

	int  row = 0;
	int  col = 0;

	for( int c = 0; c < SDSIZESQ*SDSIZESQ; c++ )
	{
		gameCell[row][col].usr = user[c] -  '0';
		gameCell[row][col].nbr = grid[c] -  '0';
		gameCell[row][col].fix = give[c] == '1';
		gameCell[row][col].sho = gameCell[row][col].usr > 0;

		if( ++col >= SDSIZESQ )
		{
			++row;
			col    = 0;
		}
	}

	gameWon = ChkGame(true);
	gameRow = 0;
	gameCol = 0;
}

// ----------------------------------------------------------------------------
void SavGame(char* grid, int type)
{
	int  row = 0;
	int  col = 0;
	char nbr;

	for( int c = 0; c < SDSIZESQ*SDSIZESQ; c++ )
	{
		switch( type )
		{
		case 0: nbr = gameCell[row][col].usr;         break;
		case 1: nbr = gameCell[row][col].nbr;         break;
		case 2: nbr = gameCell[row][col].fix ? 1 : 0; break;
		}

		grid[c]    = nbr + '0';

		if( ++col >= SDSIZESQ )
		{
			++row;
			col    = 0;
		}
	}
}

// ----------------------------------------------------------------------------
void BegGame()
{
	if( gameGen ) return;

	gameGen = true;
	srand(gameSeed);

	GenGame();

	gameLevel  = 0;
	gameSolves = 0;

	// puzzle solver's scoring system
	// TRIVIAL     |   80 or less		my beginner ( 0% to   9%) ( 81*100/900=  9)
	// EASY        |   81 - 150 points	my easy     ( 9% to  18%) (162*100/900= 18)
	// MEDIUM      |  151 - 250 points	my medium   (18% to  27%) (243*100/900= 27)
	// HARD        |  251 - 400 points	my hard     (27% to  45%) (405*100/900= 45)
	// VERY HARD   |  401 - 900 points	my extreme  (45% to 100%) (900*100/900=100)
	// DIABOLICAL  |  901 or above

	#define SOLVERTOPERCENT (900/100)
	#define MAXSOLVES        2

	const int limit[] =
	{
		gameDiff[0]*SOLVERTOPERCENT,
		gameDiff[1]*SOLVERTOPERCENT,
		gameDiff[2]*SOLVERTOPERCENT,
		gameDiff[3]*SOLVERTOPERCENT,
		gameDiff[4]*SOLVERTOPERCENT,
		gameDiff[5]*SOLVERTOPERCENT
	};
	const int range[] =
	{
		 10,
		 20,
		 60,
		 80,
		120
	};

	char puzzle[SDSIZESQ*SDSIZESQ];
	bool done,   rset,  scrd, ones, nlow, nhih;
	int  solved, score, row1, col1, row2, col2;

	bool prev     = false;
	int  lastrow1 = -1;
	int  lastcol1 = -1;

	int  lwrlimit = limit[gameType-gameBeginner];
	int  uprlimit = limit[gameType-gameBeginner+1];
	int  sizlimit = uprlimit - lwrlimit;
	int  target   = lwrlimit + sizlimit*rand()/RAND_MAX;
	int  lwrtgt   = target   - sizlimit*range[gameType]/100;
	int  uprtgt   = uprlimit;
	lwrtgt        = lwrtgt   > lwrlimit ? lwrtgt : lwrlimit;
	uprtgt        = uprtgt   < uprlimit ? uprtgt : uprlimit;

	while( true )
	{
		int  c = 0;
		for( int row = 0; row < SDSIZESQ; row++ )
		{
			for( int col = 0; col < SDSIZESQ; col++ )
			{
				puzzle[c++] = gameCell[row][col].usr;
			}
		}

		solved      = SolvePuzzle(puzzle, MAXSOLVES, &score);
		gameSolves += solved;
		gameLevel   = score*100/900;

		scrd = score  >  0;
		ones = solved == 1;
		nlow = lwrtgt <= score;
		nhih = uprtgt >= score;
		done = scrd   && ones && nlow && nhih;

		if( done ) break;

		rset = false;

		if( ones ) // one solution found
		{
			if( !scrd || !nlow ) // haven't tried enough yet so keep going
				prev  =  true;
			else                 // failed only because score was too high
			{
				if( prev )
				{
					RstGameCell(lastrow1, lastcol1, true);
					prev = false;
				}
				else
					rset = true;
			}
		}
		else // 0 or multiple solutions so restore and try again
		{
			if( prev )
			{
				RstGameCell(lastrow1, lastcol1, true);
				prev = false;
			}
			else
				rset = true;
		}

		if( rset == false )
		{
			rset =  !FndGameFixd(row1, col1);

			if( rset == false ) // still have some puzzle pieces to cover up
			{
				row2 =  SDSIZESQ - row1 - 1;
				col2 =  SDSIZESQ - col1 - 1;

				if( gameCell[row1][col1].fix && gameCell[row2][col2].fix )
				{
					lastrow1 =  row1;
					lastcol1 =  col1;
					RstGameCell(row1, col1, false);
				}
				else
					rset = true; // should never happen unless FndGameFixd is buggy
			}
		}

		if( rset )
		{
			RstGame(); // reset this game to all uncovered and try again
			prev  = false;
			score = 0;
		}
	}

	gameWon = false;
	gameRow = 0;
	gameCol = 0;
	gameGen = false;
}

// ----------------------------------------------------------------------------
bool ChkGame(bool stop)
{
	int  row, col;
	bool okay = true;

	if( !stop )
	{
		for( row = 0; row < SDSIZESQ; row++ )
		{
			for( col = 0; col < SDSIZESQ; col++ )
			{
				gameErrs[row][col] = false;
			}
		}
	}

	// check that each cell has a non-zero number

	for( row = 0; row < SDSIZESQ; row++ )
	{
		for( col = 0; col < SDSIZESQ; col++ )
		{
			if( !gameCell[row][col].usr )
			{
				okay = false;
				if( stop ) break;
				gameErrs[row][col] = true;
			}
		}

		if( !okay && stop ) break;
	}

	// check that each cell has a unique number within a given row

	if( okay || !stop )
	{
		for( row = 0; row < SDSIZESQ; row++ )
		{
			for( col = 0; col < SDSIZESQ; col++ )
			{
				for( int ccol = 0; ccol < SDSIZESQ; ccol++ )
				{
					if( ccol != col )
					{
						if( gameCell[row][ccol].usr == gameCell[row][col].usr )
						{
							okay = false;
							if( stop ) break;
							gameErrs[row][ col] = true;
							gameErrs[row][ccol] = true;
						}
					}
				}

				if( !okay && stop ) break;
			}

			if( !okay && stop ) break;
		}
	}

	// check that each cell has a unique number within a given column

	if( okay || !stop )
	{
		for( col = 0; col < SDSIZESQ; col++ )
		{
			for( row = 0; row < SDSIZESQ; row++ )
			{
				for( int crow = 0; crow < SDSIZESQ; crow++ )
				{
					if( crow != row )
					{
						if( gameCell[crow][col].usr == gameCell[row][col].usr )
						{
							okay = false;
							if( stop ) break;
							gameErrs[ row][col] = true;
							gameErrs[crow][col] = true;
						}
					}
				}

				if( !okay && stop ) break;
			}

			if( !okay && stop ) break;
		}
	}

	// check that each cell has a unique number within a given matrix

	if( okay || !stop )
	{
		for( int rtop = 0; rtop < SDSIZESQ; rtop += SDSIZE )
		{
			for( int clft = 0; clft < SDSIZESQ; clft += SDSIZE )
			{
				for( int crow = 0; crow < SDSIZE; crow++ )
				{
					for( int ccol = 0; ccol < SDSIZE; ccol++ )
					{
						for( int trow = 0; trow < SDSIZE; trow++ )
						{
							for( int tcol = 0; tcol < SDSIZE; tcol++ )
							{
								if( trow != crow || tcol != ccol )
								{
									if( gameCell[rtop+trow][clft+tcol].usr == gameCell[rtop+crow][clft+ccol].usr )
									{
										okay = false;
										if( stop ) break;
										gameErrs[rtop+crow][clft+ccol] = true;
										gameErrs[rtop+trow][clft+tcol] = true;
									}
								}
							}

							if( !okay && stop ) break;
						}

						if( !okay && stop ) break;
					}

					if( !okay && stop ) break;
				}

				if( !okay && stop ) break;
			}

			if( !okay && stop ) break;
		}
	}

	return okay;
}

// ----------------------------------------------------------------------------
void EndGame()
{
	if( !gameWon )
	{
		for( int row = 0; row < SDSIZESQ; row++ )
		{
			for( int col = 0; col < SDSIZESQ; col++ )
			{
				gameCell[row][col].usr = gameCell[row][col].nbr;
				gameCell[row][col].sho = true;
			}
		}

		gameWon = ChkGame(true);
	}
}

// ----------------------------------------------------------------------------
int GetNextNumber(int row, int col)
{
	int  rtop = (row/SDSIZE)*SDSIZE;
	int  clft = (col/SDSIZE)*SDSIZE;
	int  nbr  =  SDSIZESQ + 1;
	bool ok   =  false;

	if( !gameCell[row][col].sho )
	{
		for( nbr = 1; nbr <= SDSIZESQ; nbr++ )
		{
			bool fnd = false;

			for( int crow = 0; crow < SDSIZE; crow++ )
			{
				for( int ccol = 0; ccol < SDSIZE; ccol++ )
				{
					if( gameCell[crow+rtop][ccol+clft].sho )
					{
						if( gameCell[crow+rtop][ccol+clft].usr == nbr )
						{
							fnd = true;
							break;
						}
					}
				}

				if( fnd ) break;
			}

			if( !fnd )
			{
				ok = true;
				break;
			}
		}
	}

	return ok ? nbr : (gameCell[row][col].usr >= SDSIZESQ ? 1 : gameCell[row][col].usr+1);
}

// ----------------------------------------------------------------------------
void GetPossibles(int row, int col, bool* possible)
{
	int crow, ccol;
	int rtop = (row/SDSIZE)*SDSIZE;
	int clft = (col/SDSIZE)*SDSIZE;

	for( int n = 0; n < SDSIZESQ; n++ )
		possible[n] = true;

	// check all the cells within the given row

	for( ccol = 0; ccol < SDSIZESQ; ccol++ )
	{
		if( gameCell[row][ccol].sho )
			possible[gameCell[row][ccol].usr-1] = false;
	}

	// check all the cells within the given column

	for( crow = 0; crow < SDSIZESQ; crow++ )
	{
		if( gameCell[crow][col].sho )
			possible[gameCell[crow][col].usr-1] = false;
	}

	// check all the cells within the same matrix

	for( crow = 0; crow < SDSIZE; crow++ )
	{
		for( ccol = 0; ccol < SDSIZE; ccol++ )
		{
			if( gameCell[rtop+crow][clft+ccol].sho )
				possible[gameCell[rtop+crow][clft+ccol].usr-1] = false;
		}
	}
}

// ----------------------------------------------------------------------------
void SetDiffDefaults(int* percents)
{
	gameDiff[0] =   2;
	gameDiff[1] =   9;
	gameDiff[2] =  18;
	gameDiff[3] =  27;
	gameDiff[4] =  45;
	gameDiff[5] = 100;

	if( percents )
	{
		for( int p = 0; p < sizeof(gameDiff)/sizeof(gameDiff[0]); p++ )
			percents[p] = gameDiff[p];
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
static void GenGame()
{
	ClrGame();

	const int prow[] =
	{
		0, 0, 0, 1, 1, 1, 2, 2, 2, 0, 0, 0, 1, 1, 1, 2, 2, 2, 0, 0, 0, 1, 1, 1, 2, 2, 2,
		3, 3, 3, 4, 4, 4, 5, 5, 5, 3, 3, 3, 4, 4, 4, 5, 5, 5, 3, 3, 3, 4, 4, 4, 5, 5, 5,
		6, 6, 6, 7, 7, 7, 8, 8, 8, 6, 6, 6, 7, 7, 7, 8, 8, 8, 6, 6, 6, 7, 7, 7, 8, 8, 8,
	};
	const int pcol[] =
	{
		0, 1, 2, 0, 1, 2, 0, 1, 2, 3, 4, 5, 3, 4, 5, 3, 4, 5, 6, 7, 8, 6, 7, 8, 6, 7, 8,
		0, 1, 2, 0, 1, 2, 0, 1, 2, 3, 4, 5, 3, 4, 5, 3, 4, 5, 6, 7, 8, 6, 7, 8, 6, 7, 8,
		0, 1, 2, 0, 1, 2, 0, 1, 2, 3, 4, 5, 3, 4, 5, 3, 4, 5, 6, 7, 8, 6, 7, 8, 6, 7, 8,
	};
	const int n = sizeof(prow)/sizeof(prow[0]);

	bool  okay  = false;

	while( !okay )
	{
		okay = true;

		for( int row = 0; row < SDSIZESQ; row++ )
		{
			for( int col = 0; col < SDSIZESQ; col++ )
			{
				gameCell[row][col].nbr = 0;
				gameCell[row][col].fix = gameCell[row][col].sho = false;
				gameCell[row][col].usr = gameCell[row][col].nbr;
			}
		}

		for( int c = 0; c < n; c++ )
		{
			int row = prow[c];
			int col = pcol[c];

			gameCell[row][col].nbr = NewCellNumber(row, col);
			gameCell[row][col].fix = gameCell[row][col].sho = true;
			gameCell[row][col].usr = gameCell[row][col].nbr;

			if( !gameCell[row][col].nbr )
			{
				okay = false;
				break;
			}
		}
	}
}

// ----------------------------------------------------------------------------
static int NewCellNumber(int row, int col)
{
	int used[SDSIZESQ*SDSIZESQ], nused = 0;
	int open[SDSIZESQ*SDSIZESQ], nopen = 0;
	int crow, ccol;
	int nbr = 0;

	gameCell[row][col].nbr = 0;

	// make a list of the numbers that can't be used by this cell

	for( crow = 0; crow < SDSIZESQ; crow++ )
	{
		if( gameCell[crow][col].nbr )
			used[nused++] = gameCell[crow][col].nbr;
	}

	for( ccol = 0; ccol < SDSIZESQ; ccol++ )
	{
		if( gameCell[row][ccol].nbr )
			used[nused++] = gameCell[row][ccol].nbr;
	}

	int rtop = (row/SDSIZE)*SDSIZE;
	int clft = (col/SDSIZE)*SDSIZE;

	for( crow = 0; crow < SDSIZE; crow++ )
	{
		for( ccol = 0; ccol < SDSIZE; ccol++ )
		{
			if( gameCell[crow+rtop][ccol+clft].nbr )
				used[nused++] = gameCell[crow+rtop][ccol+clft].nbr;
		}
	}

	// make a list of the numbers that can be used by this cell

	for( int n = 1; n < 10; n++ )
	{
		bool okay = true;

		for( int u = 0; u < nused; u++ )
		{
			if( used[u] == n )
			{
				okay = false;
				break;
			}
		}

		if( okay )
			open[nopen++] = n;
	}

	if( nopen )
	{
		nbr = rand() % nopen;
		nbr = open[nbr];
	}

	return nbr;
}

// ----------------------------------------------------------------------------
static bool FndGameFixd(int& frow, int& fcol)
{
	char fixdRows[SDSIZESQ*SDSIZESQ], fixdCols[SDSIZESQ*SDSIZESQ];
	int  nfix = 0, row2, col2;

	for( int row1 = 0; row1 < SDSIZESQ; row1++ )
	{
		for( int col1 = 0; col1 < SDSIZESQ; col1++ )
		{
			row2 = SDSIZESQ - row1 - 1;
			col2 = SDSIZESQ - col1 - 1;

			if( gameCell[row1][col1].fix && gameCell[row2][col2].fix )
			{
				fixdRows[nfix] = row1;
				fixdCols[nfix] = col1;
				nfix++;
			}
		}
	}

	if( nfix )
	{
		int fval = rand() % nfix;
		frow     = fixdRows[fval];
		fcol     = fixdCols[fval];
	}

	return nfix > 0;
}

// ----------------------------------------------------------------------------
static void RstGame()
{
	for( int row = 0; row < SDSIZESQ; row++ )
	{
		for( int col = 0; col < SDSIZESQ; col++ )
		{
			gameCell[row][col].fix = gameCell[row][col].sho = true;
			gameCell[row][col].usr = gameCell[row][col].nbr;
		}
	}
}

// ----------------------------------------------------------------------------
static void RstGameCell(int row1, int col1, bool fix)
{
	int row2 = SDSIZESQ - row1 - 1;
	int	col2 = SDSIZESQ - col1 - 1;

	gameCell[row1][col1].fix = gameCell[row1][col1].sho = fix;
	gameCell[row1][col1].usr = fix ? gameCell[row1][col1].nbr : 0;
	gameCell[row2][col2].fix = gameCell[row2][col2].sho = fix;
	gameCell[row2][col2].usr = fix ? gameCell[row2][col2].nbr : 0;
}
